package ij;
import java.awt.*;
import java.awt.event.*;
import java.util.Hashtable;
import java.io.File;
import java.applet.Applet;
import java.awt.event.*;

/**
This class installs and updates ImageJ's menus.
@see ImageJ
*/

public class Menus {

	private static final int WINDOW_MENU_ITEMS = 3;
	
	private static MenuBar mbar;
	private static CheckboxMenuItem gray8Item,gray16Item,gray32Item,
			color256Item,colorRGBItem,RGBStackItem,HSBStackItem;
	private static PopupMenu popup;

	private static ImageJ ij;
	private static Applet applet;
	private static Hashtable demoImagesTable = new Hashtable();
	private static String pluginsPath;
	private static Menu plugins;
	
	static Menu window;
	static Hashtable pluginsTable = new Hashtable();
	static Hashtable shortcuts = new Hashtable();
	static int nPlugins;
	
	Menus(ImageJ ijInstance, Applet appletInstance) {
		ij = ijInstance;
		applet = appletInstance;
	}

	void addMenuBar() {
		Menu file = new Menu("File");
		addItem(file, "New...", KeyEvent.VK_N, false);
		addItem(file, "Open...", KeyEvent.VK_O, false);
		addSubMenu(file, "Open Samples");
		addSubMenu(file, "Import");
		file.addSeparator();
		addItem(file, "Close", KeyEvent.VK_W, false);
		addItem(file, "Save",  KeyEvent.VK_S, false);
		addSubMenu(file, "Save As");
		addItem(file, "Revert", KeyEvent.VK_R,  false);
		file.addSeparator();
		addPlugInItem(file, "Page Setup...", "ij.plugin.filter.Printer(\"setup\")", 0, false);
		addPlugInItem(file, "Print...", "ij.plugin.filter.Printer(\"print\")", KeyEvent.VK_P, false);
		file.addSeparator();
		addItem(file, "Quit",  0, false);
		
		Menu edit = new Menu("Edit");
		addItem(edit, "Undo", KeyEvent.VK_Z, false);
		edit.addSeparator();
		addItem(edit, "Cut", KeyEvent.VK_X, false);
		addItem(edit, "Copy", KeyEvent.VK_C, false);
		addItem(edit, "Paste", KeyEvent.VK_V, false);
		addPlugInItem(edit, "Paste Control...", "ij.plugin.frame.PasteController", 0, false);
		addPlugInItem(edit, "Clear", "ij.plugin.filter.Filler(\"clear\")", 0, false);
		addPlugInItem(edit, "Clear Outside", "ij.plugin.filter.Filler(\"outside\")", 0, false);
		edit.addSeparator();
		addItem(edit, "Select All", KeyEvent.VK_A, false);
		addItem(edit, "Select None", KeyEvent.VK_A, true);
		addItem(edit, "Restore Selection", 0, false);
		edit.addSeparator();
		addPlugInItem(edit, "Fill", "ij.plugin.filter.Filler(\"fill\")", KeyEvent.VK_F, false);
		addPlugInItem(edit, "Draw", "ij.plugin.filter.Filler(\"draw\")", KeyEvent.VK_D, false);
		addPlugInItem(edit, "Invert", "ij.plugin.filter.Filters(\"invert\")", KeyEvent.VK_I, true);
		edit.addSeparator();
		addSubMenu(edit, "Options");
		
		Menu image = new Menu("Image");
		Menu imageType = new Menu("Type");
			gray8Item = new CheckboxMenuItem("8-bit Grayscale");
			imageType.add(gray8Item);
			gray8Item.addItemListener(ij);
			gray8Item.setState(false);
			gray16Item = new CheckboxMenuItem("16-bit Grayscale");
			imageType.add(gray16Item);
			gray16Item.addItemListener(ij);
			gray16Item.setState(false);
			gray32Item = new CheckboxMenuItem("32-bit Grayscale");
			imageType.add(gray32Item);
			gray32Item.addItemListener(ij);
			gray32Item.setState(false);
			imageType.add(new MenuItem("-"));
			
			color256Item = new CheckboxMenuItem("8-bit Color");
			imageType.add(color256Item);
			color256Item.addItemListener(ij);
			color256Item.setState(false);
			colorRGBItem = new CheckboxMenuItem("RGB Color");
			imageType.add(colorRGBItem);
			colorRGBItem.addItemListener(ij);
			colorRGBItem.setState(false);
			imageType.add(new MenuItem("-"));
			
			RGBStackItem = new CheckboxMenuItem("RGB Stack");
			imageType.add(RGBStackItem);
			RGBStackItem.addItemListener(ij);
			RGBStackItem.setState(false);
			HSBStackItem = new CheckboxMenuItem("HSB Stack");
			imageType.add(HSBStackItem);
			HSBStackItem.addItemListener(ij);
			HSBStackItem.setState(false);
			image.add(imageType);
			
		image.addSeparator();
		addSubMenu(image, "Adjust");
		addPlugInItem(image, "Show Info...", "ij.plugin.filter.Info", KeyEvent.VK_I, false);
		addSubMenu(image, "Benchmarks");
		addSubMenu(image, "Stacks");
		image.addSeparator();
		addPlugInItem(image, "Crop", "ij.plugin.filter.Resizer(\"crop\")", 0, false);
		addPlugInItem(image, "Duplicate...", "ij.plugin.filter.Duplicater", KeyEvent.VK_D, true);
		addPlugInItem(image, "Scale...", "ij.plugin.filter.Scaler", KeyEvent.VK_E, false);
		addSubMenu(image, "Rotate");
		image.addSeparator();
		addSubMenu(image, "Color Tables");
		addPlugInItem(image, "Colors...", "ij.plugin.Colors", 0, false);

		Menu process = new Menu("Process");
		addPlugInItem(process, "Smooth", "ij.plugin.filter.Filters(\"smooth\")", KeyEvent.VK_S, true);
		addPlugInItem(process, "Sharpen", "ij.plugin.filter.Filters(\"sharpen\")", 0, false);
		addPlugInItem(process, "Find Edges", "ij.plugin.filter.Filters(\"edge\")", KeyEvent.VK_F, true);
		addPlugInItem(process, "Equalize", "ij.plugin.filter.Equalizer", KeyEvent.VK_E, true);
		addSubMenu(process, "Noise");
		addSubMenu(process, "Shadows");
		addSubMenu(process, "Morphology");
		addSubMenu(process, "Math");
		process.addSeparator();
		addPlugInItem(process, "Image Calculator...", "ij.plugin.ImageCalculator", 0, false);
		addPlugInItem(process, "Subtract Background...", "ij.plugin.filter.BackgroundSubtracter", 0, false);
		addItem(process, "Repeat Command", KeyEvent.VK_R, true);
		
		Menu analyze = new Menu("Analyze");
		addPlugInItem(analyze, "Measure", "ij.plugin.filter.Analyzer", KeyEvent.VK_M, false);
		addPlugInItem(analyze, "Analyze Particles...", "ij.plugin.filter.ParticleAnalyzer", 0, false);
		addPlugInItem(analyze, "Summarize", "ij.plugin.filter.Analyzer(\"sum\")", 0, false);
		addPlugInItem(analyze, "Clear Results", "ij.plugin.filter.Analyzer(\"clear\")", 0, false);
		addPlugInItem(analyze, "Set Measurements...", "ij.plugin.filter.Analyzer(\"set\")", 0, false);
		analyze.addSeparator();
		addPlugInItem(analyze, "Set Scale...", "ij.plugin.filter.ScaleDialog", 0, false);
		addPlugInItem(analyze, "Calibrate...", "ij.plugin.filter.Calibrator", 0, false);
		addItem(analyze, "Histogram", KeyEvent.VK_H, false);
		addPlugInItem(analyze, "Plot Profile", "ij.plugin.filter.Profiler(\"plot\")", KeyEvent.VK_K, false);
		addPlugInItem(analyze, "Show LUT", "ij.plugin.filter.LutViewer", 0, false);
		addSubMenu(analyze, "Fractals");
		addSubMenu(analyze, "Gels");
		addPlugInItem(analyze, "Analyze Line Graph", "ij.plugin.filter.LineGraphAnalyzer", 0, false);

		addPluginsMenu();
		if (applet==null)
			addUserPlugins();
		
		window = new Menu("Window");
		addItem(window, "ImageJ [enter]", 0, false);
		addItem(window, "Put Behind [tab]", 0, false);
		window.addSeparator();

		Menu help = new Menu("Help");
		addItem(help, "About ImageJ...", 0, false);
		help.addSeparator();
		addSubMenu(help, "About Plugins");
		
		mbar = new MenuBar();
		mbar.add(file);
		mbar.add(edit);
		mbar.add(image);
		mbar.add(process);
		mbar.add(analyze);
		mbar.add(plugins);
		mbar.add(window);
		mbar.setHelpMenu(help);
		ij.setMenuBar(mbar);
	}
	
	void addItem(Menu menu, String label, int shortcut, boolean shift) {
		MenuItem item;
		if (shortcut==0)
			item = new MenuItem(label);
		else {
			if (shift) {
				item = new MenuItem(label, new MenuShortcut(shortcut, true));
				shortcuts.put(new Integer(shortcut+200),label);
			} else {
				item = new MenuItem(label, new MenuShortcut(shortcut));
				shortcuts.put(new Integer(shortcut),label);
			}
		}
		menu.add(item);
		item.addActionListener(ij);
	}

	void addPlugInItem(Menu menu, String label, String className, int shortcut, boolean shift) {
		pluginsTable.put(label, className);
		nPlugins++;
		addItem(menu, label, shortcut, shift);
	}

	void addSubMenu(Menu menu, String name) {
		String value;
		String key = name.toLowerCase();
		int index;
 		Menu submenu=new Menu(name);
 
		index = key.indexOf(' ');
		if (index>0)
			key = key.substring(0, index);
 		for (int count=1; count<100; count++) {
			value = Prefs.getString(key + (count/10)%10 + count%10);
			if (value==null)
				break;
			if (count==1)
				menu.add(submenu);
			if (value.equals("-"))
				submenu.addSeparator();
			else {
				index = value.lastIndexOf(',');
				if (index>0) {
					addPluginItem(submenu, value, index);
					nPlugins++;
				}
			}
		}
	}

	void addPluginItem(Menu submenu, String s, int index) {
		String label = s.substring(1,index-1);
		int shortcut = 0;
		boolean shift = false;
		if (label.endsWith("]")) {
			int openBracket = label.lastIndexOf('[');
			if (openBracket>0) {
				int c = (int)label.charAt(label.length()-2);
				if (c>=65&&c<=90) {//A-Z
					shortcut = KeyEvent.VK_A+c-65;
					shift = true;
				} else if (c>=97&&c<=122) //a-z
					shortcut = KeyEvent.VK_A+c-97;
				else if (c>=48&&c<=57) //0-9
					shortcut = KeyEvent.VK_0+c-48;
				if (shortcut!=0)
					label =  label.substring(0,openBracket);
			}
		}
		if (shortcut!=0)
			addItem(submenu,label,shortcut,shift);
		else {
			MenuItem item = new MenuItem(label);
			submenu.add(item);
			item.addActionListener(ij);
		}
		String className = s.substring(index+1,s.length());
		//if (shortcut!=0)
		//	IJ.write("|"+label+"|"+className+"|");
		pluginsTable.put(label, className);
	}

	void addPluginsMenu() {
		String value,label,className;
		int index;
		plugins = new Menu("Plugins");
		for (int count=1; count<100; count++) {
			value = Prefs.getString("plug-in" + (count/10)%10 + count%10);
			if (value==null)
				break;
			if (value.equals("-"))
				plugins.addSeparator();
			else {
				index = value.lastIndexOf(',');
				if (index>0) {
					addPluginItem(plugins, value, index);
					nPlugins++;
				}
			}
		}
	}

	void addUserPlugins() {
		String homeDir = Prefs.getHomeDir();
		if (homeDir==null)
			return;
		if (homeDir.endsWith("plugins"))
			pluginsPath = homeDir;
		else {
			String separator = Prefs.getFileSeparator();
			pluginsPath = homeDir+separator+"plugins"+separator;
		}
		File pluginsFile = new File(pluginsPath);
		String[] list = pluginsFile.list();
		if (list==null)
			return;
		plugins.addSeparator();
		//ij.util.StringSorter.sort(list);
		for (int i=0; i<list.length; i++) {
			String className = list[i];
			if (className.indexOf('_')>=0 && className.endsWith(".class")&& className.indexOf('$')<0 ) {
				className = className.substring(0, className.length()-6); 
				String label = replaceUnderscores(className);
				MenuItem item = new MenuItem(label);
				plugins.add(item);
				item.addActionListener(ij);
				pluginsTable.put(label, className);
				nPlugins++;
			}
		}
	}
	
	String replaceUnderscores(String s) {
		StringBuffer sb = new StringBuffer(s);
		for (int i=0; i<sb.length(); i++)
			if (sb.charAt(i)=='_') sb.setCharAt(i, ' ');
		String s2 = sb.toString();
		s2.trim();
		return s2;
	}
	
	void installPopupMenu(ImageJ ij) {
		String s;
		int count = 0;
		MenuItem mi;
		popup = new PopupMenu("");

		while (true) {
			count++;
			s = Prefs.getString("popup" + (count/10)%10 + count%10);
			if (s==null)
				break;
			if (s.equals("-"))
				popup.addSeparator();
			else if (!s.equals("")) {
				mi = new MenuItem(s);
				mi.addActionListener(ij);
				popup.add(mi);
			}
		}
	}

	public static MenuBar getMenuBar() {
		return mbar;
	}
		
	static final int RGB_STACK=10, HSB_STACK=11;
	
	static void updateMenus() {
	
		if (ij==null) return;
		gray8Item.setState(false);
		gray16Item.setState(false);
		gray32Item.setState(false);
		color256Item.setState(false);
		colorRGBItem.setState(false);
		RGBStackItem.setState(false);
		HSBStackItem.setState(false);
		ImagePlus imp = WindowManager.getCurrentImage();
		if (imp==null)
			return;
    	int type = imp.getType();
     	if (imp.getStackSize()>1) {
    		ImageStack stack = imp.getStack();
    		if (stack.isRGB()) type = RGB_STACK;
    		else if (stack.isHSB()) type = HSB_STACK;
    	}
    	switch (type) {
    		case ImagePlus.GRAY8:
				gray8Item.setState(true);
				break;
     		case ImagePlus.GRAY16:
				gray16Item.setState(true);
				break;
     		case ImagePlus.GRAY32:
				gray32Item.setState(true);
				break;
   			case ImagePlus.COLOR_256:
				color256Item.setState(true);
				break;
    		case ImagePlus.COLOR_RGB:
				colorRGBItem.setState(true);
				break;
    		case RGB_STACK:
				RGBStackItem.setState(true);
				break;
    		case HSB_STACK:
				HSBStackItem.setState(true);
				break;
		}
		
    	//update Window menu
    	int nItems = window.getItemCount();
    	int index = WindowManager.getCurrentIndex() + WINDOW_MENU_ITEMS;
    	for (int i=WINDOW_MENU_ITEMS; i<nItems; i++) {
			CheckboxMenuItem item = (CheckboxMenuItem)window.getItem(i);
			item.setState(i==index);
		}
	}
	
	/** Returns the path to the user plugins directory. */
	public static String getPlugInsPath() {
		return pluginsPath;
	}
        
	/** Removes the last n items from the Window menu. */
	static synchronized void trimWindowMenu(int n) {
		//if (IJ.debugMode) IJ.write("Triming " + n + " items from Window menu");
		int nItems = window.getItemCount();
		if (n<=(nItems-WINDOW_MENU_ITEMS))
			for (int i=0; i<n; i++)
				window.remove(--nItems);
	}

	/** Adds one window to the end of the Window menu. */
	static synchronized void extendWindowMenu(ImagePlus imp) {
		if (ij==null) return;
		String name = imp.getTitle();
		int size = (imp.getWidth()*imp.getHeight()*imp.getStackSize())/1024;
		switch (imp.getType()) {
			case ImagePlus.GRAY32: case ImagePlus.COLOR_RGB: // 32-bit
				size *=4;
				break;
			case ImagePlus.GRAY16:  // 16-bit
				size *= 2;
				break;
			default: // 8-bit
				;
		}
		CheckboxMenuItem item = new CheckboxMenuItem(name + " " + size + "K");
		window.add(item);
		item.addItemListener(ij);
	}
	
	public static PopupMenu getPopupMenu() {
		return popup;
	}
	
}