import ij.*;
import ij.io.*;
import ij.plugin.*;
import java.io.*;


public class Leica_SP_Reader implements PlugIn {

	public void run(String arg) {
		OpenDialog od = new OpenDialog("Open Leica SP...", arg);
		String dir = od.getDirectory();
		String name = od.getFileName();
		if (name==null)
			return;
		IJ.showStatus("Opening: " + dir + name);
		try {
			FileInfo[] fi =  getFileInfo(dir, name);
			openStacks(fi);
		} catch (IOException e) {
			String msg = e.getMessage();
			if (msg==null || msg.equals(""))
				msg = ""+e;
			IJ.showMessage("Leica SP Reader", msg);
			return;
		}
	}

	FileInfo[] getFileInfo(String directory, String name) throws IOException {
		TiffDecoder td = new TiffDecoder(directory, name);
		if (IJ.debugMode) td.enableDebugging();
		FileInfo[] info = td.getTiffInfo();
		if (info==null)
			throw new IOException("This file does not appear to be in TIFF format.");
		if (IJ.debugMode) // dump tiff tags
			IJ.write(info[0].info);
		return info;
	}
	
	void openStacks(FileInfo[] fi) throws IOException {
		if (fi[0].fileType!=FileInfo.COLOR8)
			throw new IOException("This does not appear to be  a stack of 8-bit color images.");
		int maxStacks = 4;
		ImageStack[] stacks = new ImageStack[maxStacks];
		String[] colors = {"red", "green", "blue", "gray"};
		int width = fi[0].width;
		int height = fi[0].height;
		String name = fi[0].fileName;
		
		for (int i=0; i<fi.length; i++) {
			if (fi[i].width!=width || fi[i].height!=height)
				break;
			int type = getLutType(fi[i]);
			if (stacks[type]==null)
				 stacks[type] = new ImageStack(width,height);
			FileOpener fo = new FileOpener(fi[i]);
			ImagePlus imp = fo.open(false);
			if (imp!=null)
				stacks[type].addSlice("", imp.getProcessor());
		}
		
		for (int i=0; i<stacks.length; i++) {
			if (stacks[i]!=null)
				new ImagePlus(name+" ("+colors[i]+")", stacks[i]).show();
		}
	}
	
	int getLutType(FileInfo fi) {
		int rsum=0, gsum=0, bsum=0;
		for (int i=0; i<fi.lutSize; i++)
			rsum += fi.reds[i]&255;
		for (int i=0; i<fi.lutSize; i++)
			gsum += fi.greens[i]&255;
		for (int i=0; i<fi.lutSize; i++)
			bsum += fi.blues[i]&255;
		if (rsum!=0  && gsum==0 && bsum==0)
			return 0;
		else if (rsum==0 &&gsum!=0 && bsum==0)
			return 1;
		else if (rsum==0 &&gsum==0 && bsum!=0)
			return 2;
		else
			return 3; 
	}

}



